package net.minidev.json.parser;

/*
 *    Copyright 2011 JSON-SMART authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import java.io.IOException;

public class ContentHandlerDumy implements ContentHandler {
	public static ContentHandlerDumy HANDLER = new ContentHandlerDumy();

	// @Override JDK 1.5 compatibility change
	public void startJSON() throws ParseException {
	}

	// @Override JDK 1.5 compatibility change
	public void endJSON() throws ParseException {
	}

	// @Override JDK 1.5 compatibility change
	public boolean startObject() throws ParseException, IOException {
		return false;
	}

	// @Override JDK 1.5 compatibility change
	public boolean endObject() throws ParseException {
		return false;
	}

	// @Override JDK 1.5 compatibility change
	public boolean startObjectEntry(String key) throws ParseException {
		return false;
	}

	// @Override JDK 1.5 compatibility change
	public boolean endObjectEntry() throws ParseException {
		return false;
	}

	// @Override JDK 1.5 compatibility change
	public boolean startArray() throws ParseException {
		return false;
	}

	// @Override JDK 1.5 compatibility change
	public boolean endArray() throws ParseException {
		return false;
	}

	// @Override JDK 1.5 compatibility change
	public boolean primitive(Object value) throws ParseException {
		return false;
	}
}
